#ifndef TREE_H
#define TREE_H


class TTreeNode {
public:
	inline					      TTreeNode(int n = 0, int fixed = FALSE);
	virtual						  ~TTreeNode() {}
	inline TTreeNode          *GetChild(int n) const;
   inline TTreeNode          *GetParent() const;
   inline int                 GetNumChildren() const;
	void						      SetChildNumbers(int n);
   long                       CountSubtreeNodes() const;
	BOOL                       IsDescendant(TTreeNode *anc) const;

	TTreeNode                 *Parent;
	int						      MyChildNumber;
	TArrayAsVector<TTreeNode*>	Children;
};

inline					         TTreeNode::TTreeNode(int n, int fixed):
                              Children(n,0,fixed?0:2),Parent(0),MyChildNumber(0)
{
}

inline TTreeNode             *TTreeNode::GetChild(int n) const
{  return n>=0 && n<GetNumChildren()?Children[n]:0;
}

inline TTreeNode             *TTreeNode::GetParent() const
{  return Parent;
}

inline int                    TTreeNode::GetNumChildren() const
{  return Children.GetItemsInContainer();
}



class TNodeAddress : public TArrayAsVector<short> {
public:
	inline       	TNodeAddress() : TArrayAsVector<short>(9,0,10) {}

	inline int	 	Length() const { return GetItemsInContainer(); }
	inline short	GetChildNumber(int n) const {return ItemAt(GetItemsInContainer()-1-n); }
};



template<class T>
class TTree {
public:
	inline				   TTree():Root(0),NumNodes(0) {}
							  ~TTree();

	inline   T          *GetRoot() const;
	inline   long        GetNumNodes() const;
	inline	BOOL			IsLeaf(T *e) const;
	inline	BOOL			IsRoot(T *e) const;

	inline	int			GetMyChildNumber(T *e) const;
	inline	T          *GetParent(T *e)   const;
	inline	T			  *GetChild(T *e, int n) const;
	inline	int			GetNumChildren(T *e) const;
	inline	T			  *GetSibling(T *e, int n) const;
	inline	T		     *GetLeftSibling(T *e) const;
	inline	T	   	  *GetRightSibling(T *e) const;

	T                   *GetLastNode()     const;
	T                   *GetNextNode(T *e) const;
	T                   *GetPrevNode(T *e) const;

	inline   void			InsertChild(T *par,  int childnumber, T *child);
	void                 AddChild(T *par, T *child);
	void			         ReplaceChild(T *par, int childnumber, T *newchild, BOOL del=TRUE);
	void			         RemoveChild(T *par, int childnumber, BOOL del=TRUE);

	inline   void        RemoveSubtree(T *subtree, BOOL del=TRUE);
	inline   void        ReplaceSubtree(T *subtree, T *replacement, BOOL del=TRUE);

	void                 GetNodeAddress(T *node, TNodeAddress& address) const;
	T                   *NodeAt(const TNodeAddress& address) const;
private:
	void			         DoRecursive(T *e, void pre(T *), void post(T *));

	T                   *Root;
	long                 NumNodes;
};

template <class T>
TTree<T>::~TTree()
{
	if(Root)
		RemoveChild(0, 0);
}
template <class T> inline T        *TTree<T>::GetRoot() const
{  return Root;
}
template <class T> inline long      TTree<T>::GetNumNodes() const
{  return NumNodes;
}
template <class T> inline	BOOL		TTree<T>::IsLeaf(T *e) const
{  return e->Children.IsEmpty();
}
template <class T> inline	BOOL		TTree<T>::IsRoot(T *e) const
{  return e->Parent==Root;
}
template <class T> inline	int	   TTree<T>::GetMyChildNumber(T *e) const
{  return e->MyChildNumber;
}
template <class T> inline	int	   TTree<T>::GetNumChildren(T *e) const
{  return e?e->Children.GetItemsInContainer():(Root?1:0);
}
template <class T> inline	T       *TTree<T>::GetChild(T *e, int n) const
{  return (T*)(e?
	  (n>=0 && n<e->Children.GetItemsInContainer() ? e->Children[n] : 0) :
	  (n==0?Root:0));
}
template <class T> inline	T       *TTree<T>::GetSibling(T *e, int n) const
{  return GetChild(GetParent(e), n);
}
template <class T> inline	T       *TTree<T>::GetLeftSibling(T *e) const
{  return GetSibling(e, e->MyChildNumber-1);
}
template <class T> inline	T       *TTree<T>::GetRightSibling(T *e) const
{  return GetSibling(e, e->MyChildNumber+1);
}
template <class T> inline	T       *TTree<T>::
GetParent(T *e) const
{  return (T*)e->Parent;
}
template <class T> inline  void     TTree<T>::AddChild(T *par, T *child)
{  InsertChild(par, GetNumChildren(par), child);
}
template <class T> inline  void     TTree<T>::RemoveSubtree(T *subtree, BOOL del)
{  RemoveChild(GetParent(subtree),GetMyChildNumber(subtree), del);
}
template <class T> inline  void     TTree<T>::ReplaceSubtree(T *subtree, T *replacement, BOOL del)
{  ReplaceChild(GetParent(subtree),GetMyChildNumber(subtree), replacement, del);
}
template <class T>
void              TTree<T>::ReplaceChild(T *par, int childnumber, T *child, BOOL del)
{
	if(childnumber>=0 && childnumber<GetNumChildren(par)) {
		T *oldc      = GetChild(par, childnumber);
		if(oldc) {
			NumNodes -= oldc->CountSubtreeNodes();
			oldc->Parent = 0;
			if(del) {
				DoRecursive(oldc, 0, DestroyNode);
			}
		}
		if(par) {
			par->Children[childnumber] = child;
		} else if(!childnumber) {
			Root = child;
		}
		if(child) {
			child->Parent   = par; child->MyChildNumber = childnumber;
			NumNodes    += child->CountSubtreeNodes();
		}
	}
}

template <class T>
void              TTree<T>::InsertChild(T *par, int childnumber, T *child)
{
	if(par) {
		par->Children.AddAt(child, childnumber);
		par->SetChildNumbers(childnumber);
	} else if(!childnumber && !Root) {
		Root = child;
	} else
		return;
	if(child) {
		NumNodes             += child->CountSubtreeNodes();
		child->MyChildNumber  = childnumber;
		child->Parent         = par;
	}
}


template <class T>
void              TTree<T>::RemoveChild(T *e, int n, BOOL del)
{
	ReplaceChild(e, n, 0, del);
	if(e) {
		e->Children.RemoveEntry(n); e->SetChildNumbers(n);
	} else if(!n) {
		Root = 0;
	}
}


template <class T>
void              DestroyNode(T *e)
{
	delete e;
}


template <class T>
void              TTree<T>::DoRecursive(T *e, void pre(T*), void post(T*))
{
	unsigned   i, n = GetNumChildren(e);
	T         *c;

	if(pre) pre(e);
	for(i = 0; i<n; i++) {
		if(c = GetChild(e, i))
			DoRecursive(c, pre, post);
	}
	if(post) post(e);
}

template <class T>
void              TTree<T>::GetNodeAddress(T *node, TNodeAddress& address) const
{
	address.Flush();
	while(node->Parent) {
		address.Add((short)(node->MyChildNumber));
		node = GetParent(node);
	}
}

template <class T>
T                *TTree<T>::NodeAt(const TNodeAddress& address) const
{
	T    *node  = Root;

	for(int i = (int)address.GetItemsInContainer()-1; i>=0; i--) {
		node = GetChild(node, address[i]);
	}
	return node;
}

template <class T>
T                *TTree<T>::GetLastNode() const
{
	T     *lastnode = Root;
	int    cn;

	while(cn = GetNumChildren(lastnode))
		lastnode = GetChild(lastnode, cn-1);
	return lastnode;
}

template <class T>
T                *TTree<T>::GetNextNode(T *e) const
{
	int    cn = 0;
	do {
		if(cn < GetNumChildren(e)) {
			return GetChild(e, cn);
		} else {
			cn = GetMyChildNumber(e)+1;
			e  = GetParent(e);
		}
	} while(e);
	return 0;
}



template <class T>
T                *TTree<T>::GetPrevNode(T *e) const
{
	T  *prevnode;

	prevnode = GetLeftSibling(e);
	if(!prevnode) {
		prevnode = GetParent(e);
	} else {
		int nc;
		while(nc = GetNumChildren(prevnode)) {
			prevnode = GetChild(prevnode, nc-1);
		}
	}
	return prevnode;
}


#endif

